' *******************************************************************************************************
' *                                                                                                     *
' *     AiChip_JVM_Pasm_001.spin                                                                        *
' *                                                                                                     *
' *******************************************************************************************************

PRI Version
  return String("AiChip_JVM_Pasm_001")

CON
{
  
PLEASE NOTE : If modifying and re-releasing this code, please change the prefix of the source file from
              AiChip so as to avoid any confusion as to the origin of the source and update all internal
              references as appropriate. If you wish to credit the orginal source, anything  along the
              lines of, "Based on original AiChip_Jvm_XXX from AiChip Industries", is acceptable.

              AiChip is a trademark of AiChip Industries.

}
CON

' .-----------------------------------------------------------------------------------------------------.
' |     Board Configuration                                                                             |
' `-----------------------------------------------------------------------------------------------------'

  _CLKMODE      = XTAL1 + PLL16x                        ' 80MHz System Clock Speed
  _XINFREQ      = 5_000_000

OBJ

  tv            : "TV_Text"                             ' TV Text Display used for debugging
  jem           : "AiChip_JVM_JemFile_001.spin"         ' JEM Bytecode to execute ( jem.bin )
  
CON

' .-----------------------------------------------------------------------------------------------------.
' |     Hardware Configuration                                                                          |
' `-----------------------------------------------------------------------------------------------------'

  TV_PIN        = 12                                    ' Base TV Pin

CON

' .-----------------------------------------------------------------------------------------------------.
' |     Software Configuration                                                                          |
' `-----------------------------------------------------------------------------------------------------'

  NONE           = %000
  SPIN_ONLY      = %001
  PASM_ONLY      = %010
  SPIN_THEN_PASM = %110
  PASM_THEN_SPIN = %011
  
VAR

  long pcPtr                                            ' Pointer to xPc  / pc
  long spPtr                                            ' Pointer to xSp  / sp
  long bpPtr                                            ' Pointer to xBp  / bp
  long opPtr                                            ' Pointer to xOpc / jemcode
  long orPtr                                            ' Pointer to xOpr / operand
  long lhPtr                                            ' Pointer to xlhs / lhs 
  long rhPtr                                            ' Pointer to xrhs / rhs 

  long pcBase                                           ' Pointer to start of bytecode
  long spBase                                           ' Pointer to start of stack

  long debugging                                        ' Debugging flag
  long mode                                             ' Operating mode
  long event                                            ' Debugging event
  
  long stack[128]                                       ' Stack area to use 

  long xPc, xSp, xBp, xOpc, xOpr, xLhs, xRhs            ' Interface to Aichip_JVM_Pasm

  long cogDump[$200]                                    ' Copy of cog memory when debug event occurs

' *******************************************************************************************************
' *                                                                                                     *
' *     JVM Debugging                                                                                   *
' *                                                                                                     *
' *******************************************************************************************************
  
PUB StartDebugging | lastPc

  debugging := true

  tv.Start( TV_PIN )
  tv.Str( Version )

' .-----------------------------------------------------------------------------------------------------.
' |     Initialise JVM                                                                                  |
' `-----------------------------------------------------------------------------------------------------'

  Start( SPIN_THEN_PASM, jem.ByteCodeStart, @xPc, @stack )
' Start( PASM_THEN_SPIN, jem.ByteCodeStart, @xPc, @stack )
' Start( SPIN_ONLY,      jem.ByteCodeStart, @xPc, @stack )
' Start( PASM_ONLY,      jem.ByteCodeStart, @xPc, @stack )

' .-----------------------------------------------------------------------------------------------------.
' |     Main JVM Loop                                                                                   |
' `-----------------------------------------------------------------------------------------------------'

  xPc := xPc + $385
  
  repeat 5

    lastPc := xPc
    
    GetOpcodeAndOperand

    tv.Out( $0D )
    tv.Hex( lastPc, 4 )
    tv.Out( " " )
    tv.Hex( lastPc - jem.BytecodeStart, 4 )
    tv.Out( " ")

    tv.Hex( xOpc, 2 )
    tv.Out( " " )
    tv.Str( jem.OpcodeNamePtr( xOpc ) )

    if xPc > lastPc+1
      tv.Out( " " )
      if xOpr < 0
        tv.Out( "-" )
        tv.Dec( xOpr )
      else
        tv.Hex( xOpr, 4 )

    tv.Out( $0D )
    tv.Hex( xSp, 4 )
    tv.Out( " " )
    tv.hex( ( xSp - spBase ) >> 2 + 1, 2)
    ShowStack( 2 )
    ShowStack( 1 )
    ShowStack( 0 )

PRI ShowStack( depth )

    tv.out( " " )
    if ( xSp - (depth<<2) ) => spBase
      tv.Hex( long[ xSp - (depth<<2) ], 8 )
    else
      tv.Str( String("--------") )

' *******************************************************************************************************
' *                                                                                                     *
' *     Exported JVM Interface                                                                          *
' *                                                                                                     *
' *******************************************************************************************************

PUB Start( modeBits, bytecodePtr, argPtr, stackPtr )

  if mode == 0                                          ' Don't allow multiple starts
  
    mode := modeBits | $8000_0000
     
    pcBase := bytecodePtr
    spBase := stackPtr
     
    pcPtr := argPtr + ( 0 << 2 )
    spPtr := argPtr + ( 1 << 2 )
    bpPtr := argPtr + ( 2 << 2 )
    opPtr := argPtr + ( 3 << 2 )
    orPtr := argPtr + ( 4 << 2 )
    lhPtr := argPtr + ( 5 << 2 )
    rhPtr := argPtr + ( 6 << 2 )
     
    long[ pcPtr ] := pcBase
    long[ spPtr ] := spBase - 4
    long[ bpPtr ] := spBase - 4
     
    long[ @k_JemBP_Plus_4 ] := bytecodePtr + 4

    long[ $7FF0 ] := -1

    CogNew( @Jvm, @Jvm )
  
PUB GetOpcodeAndOperand | cogAdr

  Set
  Get
  repeat while event > 0
    repeat cogAdr from $000 to $1FF
      long[ $7FF0 ] := 0
      repeat while long[ $7FF0 ] == 0
      cogDump[ cogAdr ] := long[ $7FEC ]
    if debugging
      tv.Out( $0D )
      case event
        1     : IsDebug                                   ' 1 Debug
        2     : IsDebug                                   ' 2 Stop
                tv.Str( String($0D,"Stop") )
                repeat
        3     : IsSingleStep                              ' 3 SingleStep
        4     : IsNotImplemented                          ' 4 NotImplemented
        other : tv.Str( String("Unknown ") )              ' ? Unknown
                tv.Hex( event, 8 )
                repeat
    long[ $7FEC ] := 0                                  ' Let the VM execute another instruction
    long[ $7FF0 ] := 0
    Get

PRI Set

  long[ $7FE0 ] := long[ pcPtr]
  long[ $7FE4 ] := long[ spPtr]
  long[ $7FE8 ] := long[ bpPtr]

  long[ $7FF0 ] := 0

PRI Get

  repeat while ( event := long[ $7FF0 ] ) == 0

  if event < 0
  
    long[ pcPtr ] := long [ $7FE0 ]
    long[ spPtr ] := long [ $7FE4 ]
    long[ bpPtr ] := long [ $7FE8 ]
    long[ opPtr ] := long [ $7FEC ]
    long[ orPtr ] := long [ $7FF4 ]
    long[ lhPtr ] := long [ $7FF8 ]
    long[ rhPtr ] := long [ $7FFC ]
        
PRI IsDebug

  tv.Out( "+")
  tv.Hex( ( Cog(@Debug_Ret) & $1FF ) - 1, 3 )
  tv.Out( " " )
  tv.Hex( long[ $7FF4 ], 8 )
  tv.Out( " " )
  tv.Hex( long[ $7FF8 ], 8 )
  tv.Out( " " )
  tv.Hex( long[ $7FFC ], 8 )
  PauseMs( 1000 )
  
PRI IsSingleStep | yPc, yOp

  yPc := Cog(@pc) - 1
  yOp := byte[ yPc ]

  tv.Out( $0D )
  tv.Hex( yPc, 4 )
  tv.Out( "*" )
  tv.Hex( yPc - jem.BytecodeStart, 4 )
  tv.Out( " ")

  tv.Hex( yOp, 2 )
  tv.Out( " " )
  tv.Str( jem.OpcodeNamePtr( yOp ) )

PRI IsNotImplemented

  tv.Str( String("Not Implemented before $") )
  tv.Hex( Cog(@LMM_pc), 4 )
  repeat

PRI Cog( AtCogAddress )
  result := cogDump[ ( AtCogAddress - @Jvm ) >> 2 ]

PRI PauseMs( mS )
  waitcnt( CLKFREQ / 1000 * mS + CNT )

DAT

' *******************************************************************************************************
' *                                                                                                     *
' *             JVM Interpreter                                                                         *
' *                                                                                                     *
' *******************************************************************************************************

Jvm             org     $000

                add     k_LmmBp_Minus_4,PAR
                jmp     #WaitForSpin

' *******************************************************************************************************
' *                                                                                                     *
' *             Debugging Utilities                                                                     *
' *                                                                                                     *
' *******************************************************************************************************
'
' Debugging uses the top 20 bytes ( 5 longs ) of Ram ( $7FEC to $7FFF ) to communicate with the Spin
' wrapper.
' 
' These routines allow the cog to inform the Spin wrapper what's going on within itself. In all cases 
' the long[$7FF0] is set with a flag to indicate what internal 'event' is being generated. Then, once
' long[$7FF0] has been cleared, the entire $200 longs of cog are passed to the Spin program a long at 
' a time. This is done by putting the cog value in long[$7FEC] and setting long[$7FF0]. Every time the
' long[$7FF0] is cleared by the Spin wrapper the next long is passed out. After $200 have been passed
' over the code here waits for long[$7FF0] to be cleared again and execution continues. It is up to the
' Spin wrapper to not allow the cog to continue after a 'Stop' or 'NotImplemented' event.
'
' Note that during the copy we load 'dbgTmp' then wrlong that rather than altering the destination of
' the wrlong itself; this is so we can read the SPR's which cannot be used as destination registers. It
' also helps with the counting method used.
'
' In addition, writes to long[$7FF4], long[$7FF8] and long[$7FFC] can be used to pass information back
' to the debugger prior to calling any of the debugging routines. These are displayed on Debug and Stop
' calls. The Spin wrapper dictates how the events are handled and what is displayed.
'
' Example ...
'
'               wrlong  lhs,k_0000_7FF4
'               wrlong  rhs,k_0000_7FF8
'               wrlong  acc,k_0000_7FFC
'               call    #Debug
'
' Note we have to protect the Z flag so we don't screw any Cog code which relies on it being unchanged
' over calls to Debug etc. We do that by using the lsb of the destination register of the Debug_Ret
' which is otherwise unused. It saves a long. We don't need to protect C because we don't use WC in the
' debugger. 

NotImplemented  add    dbgTmp,#1                        ' 4
SingleStep      add    dbgTmp,#1                        ' 3
Stop            add    dbgTmp,#1                        ' 2
Debug           add    dbgTmp,#1                        ' 1

                wrlong  dbgTmp,k_0000_7FF0

                muxnz   Debug_Ret,k_0000_0200           ' Protect Z flag

:Dump           call    #DebugWait
:Src            mov     dbgTmp,$000                     ' No need to pre-initialise this
                wrlong  dbgTmp,k_0000_7FEC
                add     :Src,#1
                and     :Src,#$1FF WZ, NR               ' If we hit $200 we corrupted the opcode
        IF_Z    sub     :Src,k_0000_0200                ' So set it to $000 ...
                wrlong  k_0000_7FF0,k_0000_7FF0
        IF_NZ   jmp     #:Dump                          ' ... and no more to send out

                call    #DebugWait                      ' leaves dbgTmp zeroed on exit

                and     Debug_Ret,k_0000_0200 WZ, NR    ' Restore Z flag
                
NotImplemented_Ret
SingleStep_Ret
Stop_Ret
Debug_Ret       ret

DebugWait       rdlong  dbgTmp,k_0000_7FF0
                tjnz    dbgTmp,#DebugWait
DebugWait_Ret   ret

dbgTmp          long    0
                
' .-----------------------------------------------------------------------------------------------------.
' |             Push the result of an arithmetic operation ( in lhs ) to stack and execute next opcode  |
' `-----------------------------------------------------------------------------------------------------'

PushLhsThenFetch

                call    #PushLhs

' .-----------------------------------------------------------------------------------------------------.
' |             Fetch a JVM / JEM Bytcode Opcode and execute it                                         |
' `-----------------------------------------------------------------------------------------------------'

Fetch           rdbyte  opc,pc

                mov     acc,opc
                shr     acc,#1
                add     acc,#Opc_Table
                movs    :Src,acc

                add     pc,#1
                
:Src            mov     LMM_pc,0-0

                shr     opc,#1 WC, NR
        IF_C    shr     LMM_pc,#16

                and     LMM_pc,k_0000_FFFC WZ, NR

' ====  IF_NZ   call    #SingleStep
        
        IF_NZ   and     LMM_pc,k_0000_FFFC
        IF_NZ   jmp     #LMM_AddX_Fetch
        
PassToSpin      and     LMM_pc,#3
                add     LMM_pc,#Opc_Type
                jmp     LMM_pc

Opc_Type        jmp     #OprNone
                jmp     #OprByte
                jmp     #OprWord
                jmp     #OprWordSigned

OprByte         rdbyte  lhs,pc
                add     pc,#1
                jmp     #OprNone
                                
OprWord         call    #FetchWordAcc
                jmp     #OprNone

OprWordSigned   call    #FetchWordSignedAcc
                
OprNone

' .-----------------------------------------------------------------------------------------------------.
' |             Pass control to Spin to handle                                                          |
' `-----------------------------------------------------------------------------------------------------'

HandleWithSpin  wrlong  pc,k_0000_7FE0
                wrlong  sp,k_0000_7FE4
                wrlong  bp,k_0000_7FE8
                
                wrlong  opc,k_0000_7FEC
                wrlong  acc,k_0000_7FF4
                wrlong  lhs,k_0000_7FF8
                wrlong  rhs,k_0000_7FFC

                wrlong  k_FFFF_FFFF,k_0000_7FF0
                
WaitForSpin     rdlong  acc,k_0000_7FF0
                tjnz    acc,#WaitForSpin

                rdlong  pc,k_0000_7FE0
                rdlong  sp,k_0000_7FE4
                rdlong  bp,k_0000_7FE8

                jmp     #Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             Handle branching ( or not as the case may be )                                          |
' `-----------------------------------------------------------------------------------------------------'

Branch          call    #FetchWordAcc
                shl     acc,#16
                sar     acc,#16
                add     pc,acc
                sub     pc,#3
                jmp     #Fetch

SkipBranch      add     pc,#2
                jmp     #Fetch
                
' .-----------------------------------------------------------------------------------------------------.
' |             Stack handling                                                                          |
' `-----------------------------------------------------------------------------------------------------'

PopRhsThenLhs   call    #PopRhs
                call    #PopLhs
PopRhsThenLhs_ret
                ret

PopRhs          rdlong  rhs,sp
                sub     sp,#4
PopRhs_ret      ret

PopLhs          rdlong  lhs,sp
                sub     sp,#4
PopLhs_Ret      ret

PushZero        mov     rhs,#0
PushRhs         add     sp,#4
                wrlong  rhs,sp
PushRhs_ret
PushZero_ret    ret

PushLhs         add     sp,#4
                wrlong  lhs,sp
PushLhs_ret     ret

' .-----------------------------------------------------------------------------------------------------.
' |                                                                                                     |
' `-----------------------------------------------------------------------------------------------------'

CmpLhs          call    #PopLhs
                cmps    lhs,#0 WC, WZ, NR
CmpLhs_Ret      ret

CmpLhsWithRhs   call    #PopRhsThenLhs
                cmps    lhs,rhs WC, WZ, NR
CmpLhsWithRhs_Ret
                ret

' .-----------------------------------------------------------------------------------------------------.
' |                                                                                                     |
' `-----------------------------------------------------------------------------------------------------'

FetchWordAcc    rdbyte  lhs,pc
                add     pc,#1
                mov     acc,lhs

                rdbyte  rhs,pc
                add     pc,#1

                shl     acc,#8
                or      acc,rhs

FetchWordAcc_Ret
                ret

FetchWordSignedAcc
                call    #FetchWordAcc
                
                shl     acc,#16
                sar     acc,#16

FetchWordSignedAcc_Ret
                ret
                
' .-----------------------------------------------------------------------------------------------------.
' |             Opcode dispatch and operand type table                                                  |
' `-----------------------------------------------------------------------------------------------------'

'               Remove the "0*" when the opcode is supported within PASM

Opc_Table       word 0* @Opc_NOP                                        ' $00   NOP
                word 0* @Opc_ICONST_M1                                  ' $01   ICONST_M1
                word 0* @Opc_ICONST_0                                   ' $02   ICONST_0
                word 0* @Opc_ICONST_1                                   ' $03   ICONST_1
                word 0* @Opc_ICONST_2                                   ' $04   ICONST_2
                word 0* @Opc_ICONST_3                                   ' $05   ICONST_3
                word 0* @Opc_ICONST_4                                   ' $06   ICONST_4
                word 0* @Opc_ICONST_5                                   ' $07   ICONST_5
                word 0* @Opc_BIPUSH         + jem#Jem_OPR_BYTE          ' $08   BIPUSH
                word 0* @Opc_SIPUSH         + jem#Jem_OPR_SIGNED_WORD   ' $09   SIPUSH
                word 0* @Opc_ILOAD          + jem#Jem_OPR_BYTE          ' $0A   ILOAD
                word 0* @Opc_ILOAD_0                                    ' $0B   ILOAD_0
                word 0* @Opc_ILOAD_1                                    ' $0C   ILOAD_1
                word 0* @Opc_ILOAD_2                                    ' $0D   ILOAD_2
                word 0* @Opc_ILOAD_3                                    ' $0E   ILOAD_3
                word 0* @Opc_ISTORE         + jem#Jem_OPR_BYTE          ' $0F   ISTORE
                word 0* @Opc_ISTORE_0                                   ' $10   ISTORE_0
                word 0* @Opc_ISTORE_1                                   ' $11   ISTORE_1
                word 0* @Opc_ISTORE_2                                   ' $12   ISTORE_2
                word 0* @Opc_ISTORE_3                                   ' $13   ISTORE_3
                word 0* @Opc_IFEQ           + jem#Jem_OPR_BRANCH        ' $14   IFEQ
                word 0* @Opc_IFNE           + jem#Jem_OPR_BRANCH        ' $15   IFNE
                word 0* @Opc_IFLT           + jem#Jem_OPR_BRANCH        ' $16   IFLT
                word 0* @Opc_IFLE           + jem#Jem_OPR_BRANCH        ' $17   IFLE
                word 0* @Opc_IFGT           + jem#Jem_OPR_BRANCH        ' $18   IFGT
                word 0* @Opc_IFGE           + jem#Jem_OPR_BRANCH        ' $19   IFGE
                word 0* @Opc_IF_ICMPEQ      + jem#Jem_OPR_BRANCH        ' $1A   IF_ICMPEQ
                word 0* @Opc_IF_ICMPNE      + jem#Jem_OPR_BRANCH        ' $1B   IF_ICMPNE
                word 0* @Opc_IF_ICMPLT      + jem#Jem_OPR_BRANCH        ' $1C   IF_ICMPLT
                word 0* @Opc_IF_ICMPGE      + jem#Jem_OPR_BRANCH        ' $1D   IF_ICMPGE
                word 0* @Opc_IF_ICMPGT      + jem#Jem_OPR_BRANCH        ' $1E   IF_ICMPGT
                word 0* @Opc_IF_ICMPLE      + jem#Jem_OPR_BRANCH        ' $1F   IF_ICMPLE
                word 0* @Opc_IF_ACMPEQ      + jem#Jem_OPR_BRANCH        ' $20   IF_ACMPEQ
                word 0* @Opc_IF_ACMPNE      + jem#Jem_OPR_BRANCH        ' $21   IF_ACMPNE
                word 0* @Opc_IRETURN                                    ' $22   IRETURN
                word 0* @Opc_RETURN                                     ' $23   RETURN
                word 0* @Opc_INVOKEVIRTUAL  + jem#Jem_OPR_WORD          ' $24   INVOKEVIRTUAL
                word 0* @Opc_INVOKESTATIC   + jem#Jem_OPR_WORD          ' $25   INVOKESTATIC
                word 0* @Opc_GETSTATIC      + jem#Jem_OPR_WORD          ' $26   GETSTATIC
                word 0* @Opc_PUTSTATIC      + jem#Jem_OPR_WORD          ' $27   PUTSTATIC
                word 0* @Opc_POP                                        ' $28   POP
                word 0* @Opc_POP2                                       ' $29   POP2
                word 0* @Opc_DUP                                        ' $2A   DUP
                word 0* @Opc_DUP_X1                                     ' $2B   DUP_X1
                word 0* @Opc_DUP_X2                                     ' $2C   DUP_X2
                word 0* @Opc_DUP2                                       ' $2D   DUP2
                word 0* @Opc_DUP2_X1                                    ' $2E   DUP2_X1
                word 0* @Opc_DUP2_X2                                    ' $2F   DUP2_X2
                word 0* @Opc_SWAP                                       ' $30   SWAP
                word 0* @Opc_IADD                                       ' $31   IADD
                word 0* @Opc_ISUB                                       ' $32   ISUB
                word 0* @Opc_INEG                                       ' $33   INEG
                word    0
                word 0* @Opc_IOR                                        ' $35   IOR
                word    0
                word 0* @Opc_ISHL                                       ' $37   ISHL
                word    0
                word 0* @Opc_ISHR                                       ' $39   ISHR
                word    0
                word 0* @Opc_IUSHR                                      ' $3B   IUSHR
                word    0
                word 0* @Opc_IAND                                       ' $3D   IAND
                word    0
                word 0* @Opc_IXOR                                       ' $3F   IXOR
                word    0
                word 0* @Opc_I2B                                        ' $41   I2B
                word    0
                word 0* @Opc_GOTO           + jem#Jem_OPR_BRANCH        ' $43   GOTO
                word    0
                word 0* @Opc_INVOKEINTERFACE                            ' $45   INVOKEINTERFACE
                word    0
                word 0* @Opc_ARRAYLENGTH                                ' $47   ARRAYLENGTH
                word    0
                word 0* @Opc_ATHROW                                     ' $49   ATHROW
                word    0
                word 0* @Opc_CHECKCAST      + jem#Jem_OPR_WORD          ' $4B   CHECKCAST
                word    0
                word 0* @Opc_INSTANCEOF                                 ' $4D   INSTANCEOF
                word    0
                word 0* @Opc_WIDE                                       ' $4F   WIDE
                word    0
                word 0* @Opc_IINC           + jem#Jem_OPR_TWO_BYTES     ' $51   IINC
                word    0
                word 0* @Opc_GETFIELD       + jem#Jem_OPR_WORD          ' $53   GETFIELD
                word    0
                word 0* @Opc_PUTFIELD       + jem#Jem_OPR_WORD          ' $55   PUTFIELD
                word    0
                word 0* @Opc_NEW            + jem#Jem_OPR_WORD          ' $57   NEW
                word    0
                word 0* @Opc_NEWARRAY       + jem#Jem_OPR_BYTE          ' $59   NEWARRAY
                word    0
                word 0* @Opc_ANEWARRAY      + jem#Jem_OPR_TWO_BYTES     ' $5B   ANEWARRAY
                word    0
                word 0* @Opc_IALOAD                                     ' $5D   IALOAD
                word    0
                word 0* @Opc_IASTORE                                    ' $5F   IASTORE
                word    0
                word 0* @Opc_BALOAD                                     ' $61   BALOAD
                word    0
                word 0* @Opc_BASTORE                                    ' $63   BASTORE
                word    0
                word 0* @Opc_LDC            + jem#Jem_OPR_BYTE          ' $65   LDC
                word    0
                word 0* @Opc_LDC_W                                      ' $67   LDC_W
                word    0
                word 0* @Opc_LDC2_W                                     ' $69   LDC2_W
                word    0
                word 0* @Opc_IMUL                                       ' $6B   IMUL
                word    0
                word 0* @Opc_IDIV                                       ' $6D   IDIV
                word    0
                word 0* @Opc_IREM                                       ' $6F   IREM
                word    0
                word 0* @Opc_JSR            + jem#Jem_OPR_BRANCH        ' $71   JSR
                word    0
                word 0* @Opc_RET            + jem#Jem_OPR_BYTE          ' $73   RET
                word    0
                word 0* @Opc_TABLESWITCH                                ' $75   TABLESWITCH
                word    0
                word 0* @Opc_LOOKUPSWITCH                               ' $77   LOOKUPSWITCH

' .-----------------------------------------------------------------------------------------------------.
' |             JVM Variables                                                                           |
' `-----------------------------------------------------------------------------------------------------'

pc              long    0
sp              long    0
bp              long    0

opc             long    0
lhs             long    0
rhs             long    0
acc             long    0

signBit         long    0

' .-----------------------------------------------------------------------------------------------------.
' |             JVM Constants                                                                           |
' `-----------------------------------------------------------------------------------------------------'

k_0000_7FE0     long    $0000_7FE0
k_0000_7FE4     long    $0000_7FE4
k_0000_7FE8     long    $0000_7FE8
k_0000_7FEC     long    $0000_7FEC

k_0000_7FF0     long    $0000_7FF0
k_0000_7FF4     long    $0000_7FF4
k_0000_7FF8     long    $0000_7FF8
k_0000_7FFC     long    $0000_7FFC

k_0000_FFFC     long    $0000_FFFC
k_FFFF_FFFF     long    $FFFF_FFFF

DAT

' *******************************************************************************************************
' *                                                                                                     *
' *             LMM Interpreter                                                                         *
' *                                                                                                     *
' *******************************************************************************************************

CON
                        '  opcode zcri cccc ddddddddd sssssssss
 LMM                    = %010111_0001_1111_000000000_000000000                 ' jmp #<src>
 LMM_Short_Skp          = %100000_0011_1111_000000000_000000000 - 4             ' add LMM_pc,#<offset-1>*4
 LMM_Short_Fwd          = %100000_0011_1111_000000000_000000000 - 4             ' add LMM_pc,#<offset-1>*4
 LMM_Short_Rev          = %100001_0011_1111_000000000_000000000 + 4             ' sub LMM_pc,#<offset+1>*4

 LMM_IF_A               =             %1110 << 18                               ' IF_BE
 LMM_IF_AE              =             %1100 << 18                               ' IF_B
 LMM_IF_ALWAYS          =             %0000 << 18                               ' IF_NEVER
 LMM_IF_B               =             %0011 << 18                               ' IF_AE
 LMM_IF_BE              =             %0001 << 18                               ' IF_A
 LMM_IF_C               =             %0011 << 18                               ' IF_NC
 LMM_IF_C_AND_NZ        =             %1011 << 18                               ' IF_NC_OR_Z
 LMM_IF_C_AND_Z         =             %0111 << 18                               ' IF_NC_OR_NZ
 LMM_IF_C_EQ_Z          =             %0110 << 18                               ' IF_C_NE_Z
 LMM_IF_C_NE_Z          =             %1001 << 18                               ' IF_C_EQ_Z
 LMM_IF_C_OR_NZ         =             %0010 << 18                               ' IF_NC_AND_Z
 LMM_IF_C_OR_Z          =             %0001 << 18                               ' IF_NC_AND_NZ
 LMM_IF_E               =             %0101 << 18                               ' IF_NE
 LMM_IF_NC              =             %1100 << 18                               ' IF_C
 LMM_IF_NC_AND_NZ       =             %1110 << 18                               ' IF_C_OR_Z
 LMM_IF_NC_AND_Z        =             %1101 << 18                               ' IF_C_OR_NZ
 LMM_IF_NC_OR_NZ        =             %1000 << 18                               ' IF_C_AND_Z
 LMM_IF_NC_OR_Z         =             %0100 << 18                               ' IF_C_AND_NZ
 LMM_IF_NE              =             %1010 << 18                               ' IF_E
 LMM_IF_NEVER           =             %1111 << 18                               ' IF_ALWAYS
 LMM_IF_NZ              =             %1010 << 18                               ' IF_Z
 LMM_IF_NZ_AND_C        =             %1011 << 18                               ' IF_Z_OR_NC
 LMM_IF_NZ_AND_NC       =             %1110 << 18                               ' IF_Z_OR_C
 LMM_IF_NZ_OR_C         =             %0010 << 18                               ' IF_Z_AND_NC
 LMM_IF_NZ_OR_NC        =             %1000 << 18                               ' IF_Z_AND_C
 LMM_IF_Z               =             %0101 << 18                               ' IF_NZ
 LMM_IF_Z_AND_C         =             %0111 << 18                               ' IF_NZ_OR_NC
 LMM_IF_Z_AND_NC        =             %1101 << 18                               ' IF_NZ_OR_C
 LMM_IF_Z_EQ_C          =             %0110 << 18                               ' IF_Z_NE_C
 LMM_IF_Z_NE_C          =             %1001 << 18                               ' IF_Z_EQ_C
 LMM_IF_Z_OR_C          =             %0001 << 18                               ' IF_NZ_AND_NC
 LMM_IF_Z_OR_NC         =             %0100 << 18                               ' IF_NZ_AND_C

DAT

LmmVm           org     $

' .-----------------------------------------------------------------------------------------------------.
' |             LMM Internal registers and constants                                                    |
' `-----------------------------------------------------------------------------------------------------'

k_LmmBp_Minus_4 long    -(@Jvm+4)                       ' Base adjust for @Label to absolute hub address

k_JemBP_Plus_4  long    0                               ' Absolute Hub address of JEM bytecode plus 4

k_0000_0200     long    $0000_0200                      ' ---- ---- ---- --dd dddd ddD- ---- ----

LMM_pc          long    0                               ' LMM Program Counter
LMM_sp          long    0                               ' LMM Stack Pointer ( 0 = Internal Stack )

' .-----------------------------------------------------------------------------------------------------.
' |             CONDITIONAL SKIP                        long    LMM | LMM_Bra_Skp ^ LMM_Cond | n << 9   |
' `-----------------------------------------------------------------------------------------------------'

LMM_Bra_Skp

' .-----------------------------------------------------------------------------------------------------.
' |             JMP     #$+longs                        long    LMM | LMM_Bra_Fwd | longs << 9          |
' `-----------------------------------------------------------------------------------------------------'

LMM_Bra_Fwd     mov     LMM_Bra_Neg2,LMM_Bra_Neg1       ' Double negate
                jmp     #LMM_Bra_Xxx

' .-----------------------------------------------------------------------------------------------------.
' |             JMP     #$-longs                        long    LMM | LMM_Bra_Rev | longs << 9          |
' `-----------------------------------------------------------------------------------------------------'

LMM_Bra_Rev     mov     LMM_Bra_Neg2,#0                 ' Single negate
                
LMM_Bra_Xxx     mov     acc,LMM_pc
                and     acc,#1
                tjnz    acc,#:skp
                mov     LMM_Opc1,LMM_Opc2

:Skp            shr     LMM_Opc1,#9
                and     LMM_Opc1,#$1FF
                shl     LMM_Opc1,#2

LMM_Bra_Neg1    neg     LMM_Opc1,LMM_Opc1
LMM_Bra_Neg2    long    0-0

                andn    LMM_pc,#3
                sub     LMM_pc,#4
                add     LMM_pc,LMM_Opc1

                jmp     #LMM_Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             JMP     #@Label                         long    LMM | LMM_Jmp                           |
' |                                                     long    @Label                                  |
' `-----------------------------------------------------------------------------------------------------'

LMM_Add4_Jmp    add     LMM_pc,#4

LMM_Jmp         rdlong  LMM_pc,LMM_pc

' .-----------------------------------------------------------------------------------------------------.
' |             LMM Fetch and Execute Loop                                                              |
' `-----------------------------------------------------------------------------------------------------'

LMM_AddX_Fetch  add     LMM_pc,k_LmmBp_Minus_4          ' Adjust @Label address to actual hub addresses

LMM_Add4_Fetch  add     LMM_pc,#4

LMM_Fetch       andn    LMM_pc,#3                       ' Ensure correct long alignment

LMM_Loop        rdlong  LMM_Opc1,LMM_pc
                add     LMM_pc,#7
LMM_Opc1        nop

                rdlong  LMM_Opc2,LMM_pc
                add     LMM_pc,#1
LMM_Opc2        nop

                jmp     #LMM_Loop

' .-----------------------------------------------------------------------------------------------------.
' |             LOAD    reg,#longConstant               long    LMM | LMM_Load | reg << 9               |
' |                                                     long    longConstant                            |
' `-----------------------------------------------------------------------------------------------------'

LMM_load        mov     acc,LMM_pc
                and     acc,#1
                tjnz    acc,#:Skp
                mov     LMM_Opc1,LMM_Opc2
:Skp            shr     LMM_Opc1,#9
                movd    :Dst,LMM_Opc1
                nop
:Dst            rdlong   0-0,LMM_pc
                jmp     #LMM_Add4_Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             GTI     dreg,sreg                       long    LMM | LMM_Gti | dreg << 9               |
' |             MOV     dreg,[sreg]                     long    sreg                                    |
' `-----------------------------------------------------------------------------------------------------'

LMM_Gti         mov     acc,LMM_pc
                and     acc,#1
                tjnz    acc,#:Skp
                mov     LMM_Opc1,LMM_Opc2
:Skp            shr     LMM_Opc1,#9
                rdlong  acc,LMM_pc
                movs    :Src,acc
                movd    :DstSrc,Lmm_Opc1
:Src            movs    :DstSrc,0-0
                add     LMM_pc,#4
:DstSrc         mov     0-0,0-0
                jmp     #LMM_Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             STI     sreg,dreg                       long    LMM | LMM_Sti | sreg << 9               |
' |             MOV     [dreg],sreg                     long    dreg                                    |
' `-----------------------------------------------------------------------------------------------------'

LMM_Sti         mov     acc,LMM_pc
                and     acc,#1
                tjnz    acc,#:Skp
                mov     LMM_Opc1,LMM_Opc2
:Skp            shr     LMM_Opc1,#9
                rdlong  acc,LMM_pc
                movd    :Src,acc
                movs    :DstSrc,LMM_Opc1
:Src            movd    :DstSrc,0-0
                add     LMM_pc,#4
:DstSrc         mov     0-0,0-0
                jmp     #LMM_Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             EXECUTE reg                             long    LMM | LMM_Execute | reg << 9            |
' `-----------------------------------------------------------------------------------------------------'

LMM_Execute     mov     acc,LMM_pc
                and     acc,#1
                tjnz    acc,#:Skp
                mov     LMM_Opc1,LMM_Opc2
:Skp            shr     LMM_Opc1,#9
                movs    :Src,LMM_Opc1
                nop
:Src            mov     :Opc,0-0
                nop
:Opc            long    0-0
                jmp     #LMM_Fetch
                
' .-----------------------------------------------------------------------------------------------------.
' |             JMP     reg                             long    LMM | LMM_Jmp_Reg | reg << 9            |
' `-----------------------------------------------------------------------------------------------------'

LMM_Jmp_Reg     mov     acc,LMM_pc
                and     acc,#1
                tjnz    acc,#:Skp
                mov     LMM_Opc1,LMM_Opc2
:Skp            shr     LMM_Opc1,#9
                movs    :Src,LMM_Opc1
                nop
:Src            mov     LMM_pc,0-0
                jmp     #LMM_AddX_Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             CASE     reg                            long    LMM | LMM_Case | reg << 9               |
' |                                                     word    @Label0                                 |
' |                                                     word    @Label1                                 |
' |                                                     :                                               |
' |                                                     word    @LabelN                                 |
' `-----------------------------------------------------------------------------------------------------'

LMM_Case        mov     acc,LMM_pc
                and     acc,#1
                tjnz    acc,#:Skp
                mov     LMM_Opc1,LMM_Opc2
:Skp            shr     LMM_Opc1,#9

                movs    :Src1,LMM_Opc1
                movs    :Src2,LMM_Opc1
                andn    LMM_pc,#3                       ' Ensure pc is aligned correctly
:Src1           add     LMM_pc,0-0
:Src2           add     LMM_pc,0-0

                rdword  LMM_pc,LMM_pc
                jmp     #LMM_AddX_Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             CALL    #@Label                         long    LMM | LMM_Call                          |
' |                                                     long    @Label                                  |
' `-----------------------------------------------------------------------------------------------------'

LMM_Call        tjz     LMM_sp,#LMM_Call_Int            ' Use internal stack if sp == 0

                wrword  LMM_pc,LMM_sp                   ' Push to external stack
                add     LMM_sp,#2
                jmp     #LMM_Jmp
                
LMM_Call_Int    mov     $1EF,LMM_pc                     ' Push to internal stack
                sub     LMM_Call_Int,k_0000_0200        ' Adjust push address
                sub     LMM_Ret_Int,#1                  ' Adjust pop address
                jmp     #LMM_Jmp

' .-----------------------------------------------------------------------------------------------------.
' |             RET                                     long    LMM | LMM_Ret                           |
' `-----------------------------------------------------------------------------------------------------'

LMM_Ret         tjz     LMM_sp,#LMM_Ret_Int             ' Use internal stack if sp == 0

                sub     LMM_sp,#2                       ' Pop from external stack
                rdword  LMM_pc,LMM_sp
                jmp     #LMM_Add4_Fetch
                                                        
LMM_Ret_Int     mov     LMM_pc,$1F0                     ' Pop from internal stack
                add     LMM_Call_Int,k_0000_0200        ' Adjust push address
                add     LMM_Ret_Int,#1                  ' Adjust pop address
                jmp     #LMM_Add4_Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             DJNZ    reg,#@Label                     long    LMM | LMM_Djnz | reg << 9               |
' |                                                     long    @Label                                  |
' `-----------------------------------------------------------------------------------------------------'

LMM_Djnz        movi    LMM_Djnz_OpcDst,#%111001_00_1   ' Djnz opcode

LMM_Djnz_Xxx    mov     acc,LMM_pc
                and     acc,#1
                tjnz    acc,#:Skp
                mov     LMM_Opc1,LMM_Opc2
:Skp            shr     LMM_Opc1,#9
                movd    LMM_Djnz_OpcDst,LMM_Opc1
                nop     
LMM_Djnz_OpcDst djnz    0-0,#LMM_Jmp
                jmp     #LMM_Add4_Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             TJZ     reg,#@Label                     long    LMM | LMM_Tjz | reg << 9                |
' |                                                     long    @Label                                  |
' `-----------------------------------------------------------------------------------------------------'

LMM_Tjz         movi    LMM_Djnz_OpcDst,#%111011_00_0   ' Tjz opcode
                jmp     #LMM_Djnz_Xxx

' .-----------------------------------------------------------------------------------------------------.
' |             TJNZ    reg,#@Label                     long    LMM | LMM_Tjnz | reg << 9               |
' |                                                     long    @Label                                  |
' `-----------------------------------------------------------------------------------------------------'

LMM_Tjnz        movi    LMM_Djnz_OpcDst,#%111010_00_0   ' Tjnz Opcode
                jmp     #LMM_Djnz_Xxx

' *******************************************************************************************************
' *                                                                                                     *
' *             End of LMM Interpreter                                                                  *
' *                                                                                                     *
' *******************************************************************************************************

DAT

' *******************************************************************************************************
' *                                                                                                     *
' *             Place LMM Program Variables / Registers / Global Constants / Fast Subroutines here      *
' *                                                                                                     *
' *******************************************************************************************************

                fit     $1F0                            ' Ensure we do not exceed Cog maximum

DAT

' *******************************************************************************************************
' *                                                                                                     *
' *             Place LMM Program here                                                                  *
' *                                                                                                     *
' *******************************************************************************************************

LmmCodeStart

' .-----------------------------------------------------------------------------------------------------.
' |             NOP                                                                                 $00 |
' |             Meaning: Do nothing.                                                                    |
' `-----------------------------------------------------------------------------------------------------'

Opc_NOP         jmp     #Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             ICONST_M1                                                                           $01 |
' |             Meaning: Push the integer constant -1.                                                  |
' |             Description: This bytecode pushes the integer constant -1 onto the stack.               |
' |-----------------------------------------------------------------------------------------------------|
' |             ICONST_0                                                                        $02-$07 |
' |             Meaning: Push the integer constant <C>.                                                 |
' |             Description: This bytecode pushes an integer constant onto the stack. There             |
' |             are six associated bytecodes, one for each of the integers 0-5: iconst_0,               |
' |             iconst_1, iconst_2, iconst_3, iconst_4, and iconst_5.                                   |
' `-----------------------------------------------------------------------------------------------------'

Opc_ICONST_M1
Opc_ICONST_0
Opc_ICONST_1
Opc_ICONST_2
Opc_ICONST_3
Opc_ICONST_4
Opc_ICONST_5    mov     lhs,opc
                sub     lhs,#jem#Jem_ICONST_0
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             BIPUSH                                                                              $08 |
' |             Meaning: Push a 1-byte signed integer.                                                  |
' |             Description: The first byte following the opcode, byte1, is interpreted as a            |
' |             signed 8-bit value. This value is expanded to a 32-bit integer and pushed onto          |
' |             the operand stack.                                                                      |
' `-----------------------------------------------------------------------------------------------------'

Opc_BIPUSH      rdbyte  lhs,pc
                add     pc,#1
                shl     lhs,#24
                sar     lhs,#24
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             SIPUSH                                                                              $09 |
' |             Meaning: Push a 2-byte signed integer.                                                  |
' |             Description: The first and second byte following the opcode, byte1 and byte2,           |
' |             are assembled into a signed 16-bit value. This value is expanded to a 32-bit            |
' |             integer and pushed onto the operand stack.                                              |
' `-----------------------------------------------------------------------------------------------------'

Opc_SIPUSH      call    #FetchWordSignedAcc
                mov     lhs,acc
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             ILOAD                                                                               $0A |
' |             Meaning: Load integer from local variable.                                              |
' |             Description: The value of the local variable in the current Java frame is               |
' |             pushed onto the operand stack. This value must be of type integer.                      |
' |-----------------------------------------------------------------------------------------------------|
' |             ILOAD_0                                                                             $0B |
' |             ILOAD_1                                                                             $0C |
' |             ILOAD_2                                                                             $0D |
' |             ILOAD_3                                                                             $0E |
' |             Meaning: Load integer from local variable with specific index.                          |
' |             Description: The value of the local variable <l> in the current Java frame is           |
' |             pushed onto the operand stack.                                                          |
' |             This value must be of type integer. There are four of these bytecodes, one for          |
' |             each of the integers 0-3: iload_0, iload_1, iload_2, and iload_3.                       |
' `-----------------------------------------------------------------------------------------------------'

Opc_ILOAD       rdbyte  opc,pc
                add     pc,#1
                add     opc,#jem#Jem_ILOAD_0

Opc_ILOAD_0
Opc_ILOAD_1
Opc_ILOAD_2
Opc_ILOAD_3     sub     opc,#jem#Jem_ILOAD_0
                shl     opc,#2
                add     opc,bp
                rdlong  lhs,opc
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             ISTORE                                                                              $0F |
' |             Meaning: Store integer into local variable.                                             |
' |             Description: Local variable in the current Java frame is set to a value. This           |
' |             value must be an integer.                                                               |
' |-----------------------------------------------------------------------------------------------------|
' |             ISTORE_0                                                                            $10 |
' |             ISTORE_1                                                                            $11 |
' |             ISTORE_2                                                                            $12 |
' |             ISTORE_3                                                                            $13 |
' |             Meaning: Store integer into local variable with specific index.                         |
' |             Description: Local variable <l> in the current Java frame is set to a value.            |
' |             This value must be an integer. There are four of these bytecodes, one for each          |
' |             of the integers 0-3: istore_0, istore_1, istore_2, and istore_3.                        |
' `-----------------------------------------------------------------------------------------------------'


Opc_ISTORE      rdbyte  opc,pc
                add     pc,#1
                add     opc,#jem#Jem_ISTORE_0

Opc_ISTORE_0
Opc_ISTORE_1
Opc_ISTORE_2
Opc_ISTORE_3    sub     opc,#jem#Jem_ISTORE_0
                shl     opc,#2
                add     opc,bp
                wrlong  lhs,opc
                jmp     #Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             IFEQ                                                                                $14 |
' |             Meaning: Branch if equal.                                                               |
' `-----------------------------------------------------------------------------------------------------'

Opc_IFEQ        call    #CmpLhs
        IF_E    jmp     #Branch
                jmp     #SkipBranch

' .-----------------------------------------------------------------------------------------------------.
' |             IFNE                                                                                $15 |
' |             Meaning: Branch if not equal.                                                           |
' `-----------------------------------------------------------------------------------------------------'

Opc_IFNE        call    #CmpLhs
        IF_NE   jmp     #Branch
                jmp     #SkipBranch

' .-----------------------------------------------------------------------------------------------------.
' |             IFLT                                                                                $16 |
' |             Meaning: Branch if less than.                                                           |
' `-----------------------------------------------------------------------------------------------------'

Opc_IFLT        call    #CmpLhs
        IF_B    jmp     #Branch
                jmp     #SkipBranch

' .-----------------------------------------------------------------------------------------------------.
' |             IFLE                                                                                $17 |
' |             Meaning: Branch if less than or equal.                                                  |
' `-----------------------------------------------------------------------------------------------------'

Opc_IFLE        call    #CmpLhs
        IF_BE   jmp     #Branch
                jmp     #SkipBranch

' .-----------------------------------------------------------------------------------------------------.
' |             IFGT                                                                                $18 |
' |             Meaning: Branch if greater than.                                                        |
' `-----------------------------------------------------------------------------------------------------'

Opc_IFGT        call    #CmpLhs
        IF_A    jmp     #Branch
                jmp     #SkipBranch

' .-----------------------------------------------------------------------------------------------------.
' |             IFGE                                                                                $19 |
' |             Meaning: Branch if greater than or equal.                                               |
' `-----------------------------------------------------------------------------------------------------'

Opc_IFGE        call    #CmpLhs
        IF_AE   jmp     #Branch
                jmp     #SkipBranch

' .-----------------------------------------------------------------------------------------------------.
' |             IF_ICMPEQ                                                                           $1A |
' |             Meaning: Branch if integers v1 and v2 are equal.                                        |
' `-----------------------------------------------------------------------------------------------------'

Opc_IF_ICMPEQ   call    #CmpLhsWithRhs
        IF_E    jmp     #Branch
                jmp     #SkipBranch

' .-----------------------------------------------------------------------------------------------------.
' |             IF_ICMPNE                                                                           $1B |
' |             Meaning: Branch if integers v1 and v2 are not equal.                                    |
' `-----------------------------------------------------------------------------------------------------'

Opc_IF_ICMPNE   call    #CmpLhsWithRhs
        IF_NE   jmp     #Branch
                jmp     #SkipBranch

' .-----------------------------------------------------------------------------------------------------.
' |             IF_ICMPLT                                                                           $1C |
' |             Meaning: Branch if integer v1 is less than integer v2.                                  |
' `-----------------------------------------------------------------------------------------------------'

Opc_IF_ICMPLT   call    #CmpLhsWithRhs
        IF_B    jmp     #Branch
                jmp     #SkipBranch

' .-----------------------------------------------------------------------------------------------------.
' |             IF_ICMPGE                                                                           $1D |
' |             Meaning: Branch if integer v1 greater than or equal to integer v2.                      |
' `-----------------------------------------------------------------------------------------------------'

Opc_IF_ICMPGE   call    #CmpLhsWithRhs
        IF_AE   jmp     #Branch
                jmp     #SkipBranch

' .-----------------------------------------------------------------------------------------------------.
' |             IF_ICMPGT                                                                           $1E |
' |             Meaning: Branch if integer v1 is greater than integer v2.                               |
' `-----------------------------------------------------------------------------------------------------'

Opc_IF_ICMPGT   call    #CmpLhsWithRhs
        IF_A    jmp     #Branch
                jmp     #SkipBranch

' .-----------------------------------------------------------------------------------------------------.
' |             IF_ICMPLE                                                                           $1F |
' |             Meaning: Branch if integer v1 less than or equal to integer v2.                         |
' `-----------------------------------------------------------------------------------------------------'

Opc_IF_ICMPLE   call    #CmpLhsWithRhs
        IF_BE   jmp     #Branch
                jmp     #SkipBranch

' .-----------------------------------------------------------------------------------------------------.
' |             IF_ACMPEQ                                                                           $20 |
' |             Meaning: Branch if object references equal.                                             |
' `-----------------------------------------------------------------------------------------------------'

Opc_IF_ACMPEQ
                jmp     #HandleWithSpin
        IF_E    jmp     #Branch
                jmp     #SkipBranch

' .-----------------------------------------------------------------------------------------------------.
' |             IF_ACMPNE                                                                           $21 |
' |             Meaning: Branch if object references not equal.                                         |
' `-----------------------------------------------------------------------------------------------------'

Opc_IF_ACMPNE
                jmp     #HandleWithSpin
        IF_NE   jmp     #Branch
                jmp     #SkipBranch

' .-----------------------------------------------------------------------------------------------------.
' |             IRETURN                                                                             $22 |
' |             Meaning: Return integer from function.                                                  |
' `-----------------------------------------------------------------------------------------------------'

Opc_IRETURN     call    #PopLhs
                sub     bp,#4
                rdlong  pc,bp
                sub     bp,#4
                rdlong  sp,bp
                sub     bp,#4
                rdlong  bp,bp
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             RETURN                                                                              $23 |
' |             Meaning: Return (void) from method.                                                     |
' `-----------------------------------------------------------------------------------------------------'

Opc_RETURN      sub     bp,#4
                rdlong  pc,bp
                sub     bp,#4
                rdlong  sp,bp
                sub     bp,#4
                rdlong  bp,bp
                jmp     #Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             INVOKEVIRTUAL                                                                       $24 |
' |             Meaning: Invoke method based on runtime type.                                           |
' `-----------------------------------------------------------------------------------------------------'

Opc_INVOKEVIRTUAL
                jmp     #HandleWithSpin
                jmp     #Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             INVOKESTATIC                                                                        $25 |
' |             Meaning: Invoke a class (static) method.                                                |
' `-----------------------------------------------------------------------------------------------------'

'               SP -->
'       .----.        .----.          .----.          .----.
'       | A  |        |    |          | BP |          | BP |
'       |----|        |----|          |----|          |----|
' SP -->| B  |        |    |          | SP |          | SP |
'       `----'        |----|          |----|          |----|
'                     |    |          | PC |          | PC |
'                     |----|          |----|          |----|
'                     | A  |    BP -->| A  |    BP -->| A  |
'                     |----|          |----|          |----|
'               XP -->| B  |    SP -->| B  |          | B  |
'                     `----'          `----'          |----|
'                                                     | 0  |
'                                                     |----|
'                                               SP -->| 0  |
'                                                     `----'

'  xp := sp + 3*4
'  sp := sp - params*4
'  copy parms x [xp] := [xp-3*4]
'  push bp,sp,pc      
'  bp := sp+1*4
'  sp := xp
'  push locals x 0

Opc_INVOKESTATIC         

'               call    #FetchWordAcc           ' Adr of method table
'               add     acc,k_JemBp_Plus_4
'               add     acc,k_CogBp_Minus_4
'               
'               rdbyte  lhs,acc                 ' How many parameters
'
'               mov     xp,sp                   ' xp := sp + 3*4
'               add     xp,#3*4
'                     
'               sub     sp,acc                  '  sp := sp - params*4
'               add     sp,acc
'               
'               shr     acc,#1
'
'               long    LMM | LMM_Tjz | acc << 9
'               long    @Opc_INVOKESTATIC_P_None
'                             
'               mov     yp,xp

Opc_INVOKESTATIC_P_Loop

'               sub     yp,#3*4
'               rdlong  rhs,yp
'               add     yp,#3*4
'               wrlong  rhs,yp
        
'               sub     yp,#4
                
'               long    LMM | LMM_Djnz | acc << 9
'               long    @Opc_INVOKESTATIC_P_Loop

Opc_INVOKESTATIC_P_None
                         
'               mov     yp,sp
'               add     yp,#4
'               wrlong  bp,yp
'               add     yp,#4
'               wrlong  sp,yp
'               add     yp,#4
'               wrlong  pc,yp   
                
'               add     yp,#4
'               mov     bp,yp
                
'               mov     sp,xp

'               add     lhs,#1                  ' How many locals
'               rdbyte  acc,lhs

'               add     sp,acc
'               add     sp,acc

'               add     lhs,#1                  ' Flags ( native ? )
'               rdbyte  acc,lhs

'               long    LMM | LMM_Tjz | acc << 9
'               long    @Opc_INVOKESTATIC_Native
                                   
'               add     lhs,#1                  ' Adr of method
'               mov     pc,lhs
'               call    #FetchWordLhs
'               mov     pc,lhs
'               add     pc,k_JemBp_Plus_4

'               jmp     #Fetch
                                                
Opc_INVOKESTATIC_Native
                
'               long    LMM | LMM_Jmp
'               long    @Opc_RETURN

                jmp     #HandleWithSpin

' .-----------------------------------------------------------------------------------------------------.
' |             GETSTATIC                                                                           $26 |
' |             Meaning: Get 16bit static field from class.                                             |
' `-----------------------------------------------------------------------------------------------------'

Opc_GETSTATIC
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             PUTSTATIC                                                                           $27 |
' |             Meaning: Set 16bit static field in class.                                               |
' `-----------------------------------------------------------------------------------------------------'

Opc_PUTSTATIC
                jmp     #HandleWithSpin
                jmp     #Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             POP                                                                                 $28 |
' |             Meaning: Pop the top word from the stack.                                               |
' `-----------------------------------------------------------------------------------------------------'

Opc_POP         Call    #PopLhs
                jmp     #Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             POP2                                                                                $29 |
' |             Meaning: Pop the top two words from the stack.                                          |
' `-----------------------------------------------------------------------------------------------------'

Opc_POP2        call    #PopRhsThenLhs
                jmp     #Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             DUP                                                                                 $2A |
' |             Meaning: Duplicate the top word on the stack.                                           |
' `-----------------------------------------------------------------------------------------------------'

Opc_DUP         rdlong  lhs,sp
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             DUP_X1                                                                              $2B |
' |             Meaning: Duplicate the top word on the stack and insert a copy two words down           |
' |             in the stack.                                                                           |
' `-----------------------------------------------------------------------------------------------------'

'       .---.           .---.
'       | L |           | R |
'       |---|           |---|
' SP -->| R |           | L |
'       `---'           |---|
'                 SP -->| R |
'                       `---'

Opc_DUP_X1      call    #PopRhsThenLhs
                Call    #PushRhs
                call    #PushLhs
                call    #PushRhs
                jmp     #Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             DUP_X2                                                                              $2C |
' |             Meaning: Duplicate the top word on the stack and insert the copy three words            |
' |             down in the stack.                                                                      |
' `-----------------------------------------------------------------------------------------------------'
'       .---.           .---.
'       | L |           | R |
'       |---|           |---|
'       | A |           | L |
'       |---|           |---|
' SP -->| R |           | A |
'       `---'           |---|
'                 SP -->| R |
'                       `---'

Opc_DUP_X2      call    #PopRhsThenLhs
                mov     acc,lhs
                rdlong  lhs,sp
                
                wrlong  rhs,sp
                call    #PushLhs
                mov     rhs,acc
                call    #PushRhs
                call    #PushLhs
                
                jmp     #Fetch                

' .-----------------------------------------------------------------------------------------------------.
' |             DUP2                                                                                $2D |
' |             Meaning: Duplicate the top two words on the stack.                                      |
' `-----------------------------------------------------------------------------------------------------'
'
'       .---.           .---.
'       | R |           | R |
'       |---|           |---|
' SP -->| L |           | L |
'       `---'           |---|
'                       | R |
'                       |---|
'                 SP -->| L |
'                       `---'

Opc_DUP2        rdlong  lhs,sp
                sub     sp,#4
                rdlong  rhs,sp
                add     sp,#8
                wrlong  rhs,sp
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             DUP2_X1                                                                             $2E |
' |             Meaning: Duplicate the top two words on the stack and insert the copies two             |
' |             words down in the stack.                                                                |
' `-----------------------------------------------------------------------------------------------------'
'
'       .---.           .---.
'       | L |           | A |
'       |---|           |---|
'       | A |           | R |
'       |---|           |---|
' SP -->| R |           | L |
'       `---'           |---|
'                       | A |
'                       |---|
'                 SP -->| R |
'                       `---'

Opc_DUP2_X1     call    #PopRhsThenLhs
                mov     acc,lhs
                rdlong  lhs,sp

                wrlong  acc,sp                
                call    #PushRhs
                call    #PushLhs
                add     sp,#4
                wrlong  acc,sp
                call    #PushRhs
                
                jmp     #Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             DUP2_X2                                                                             $2F |
' |             Meaning: Duplicate the top two words on the stack and insert the copies three           |
' |             words down in the stack.                                                                |
' `-----------------------------------------------------------------------------------------------------'
'
'       .---.           .---.
'       | 21|           | 11|
'       |---|           |---|
'       | 20|           | 10|
'       |---|           |---|
'       | 11|           | 21|
'       |---|           |---|
' SP -->| 10|           | 20|
'       `---'           |---|
'                       | 11|
'                       |---|
'                 SP -->| 10|
'                       `---'
Opc_DUP2_X2

                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             SWAP                                                                                $30 |
' |             Meaning: Swap the top two elements on the stack.                                        |
' `-----------------------------------------------------------------------------------------------------'

Opc_SWAP        call    #PopRhs
                rdlong  lhs,sp
                wrlong  rhs,sp
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             IADD                                                                                $31 |
' |             Meaning: Integer add.                                                                   |
' `-----------------------------------------------------------------------------------------------------'

Opc_IADD        call    #PopRhsThenLhs
                adds    lhs,rhs
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             ISUB                                                                                $32 |
' |             Meaning: Integer subtract.                                                              |
' `-----------------------------------------------------------------------------------------------------'

Opc_ISUB        call    #PopRhsThenLhs
                subs    lhs,rhs
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             INEG                                                                                $33 |
' |             Meaning: Integer negate.                                                                |
' `-----------------------------------------------------------------------------------------------------'

Opc_INEG        call    #PopLhs
                neg     lhs,lhs
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             IOR                                                                                 $34 |
' |             Meaning: Integer boolean OR.                                                            |
' `-----------------------------------------------------------------------------------------------------'

Opc_IOR         call    #PopRhsThenLhs
                or      lhs,rhs
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             ISHL                                                                                $35 |
' |             Meaning: Integer shift left.                                                            |
' `-----------------------------------------------------------------------------------------------------'

Opc_ISHL        call    #PopRhsThenLhs
                shl     lhs,rhs
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             ISHR                                                                                $36 |
' |             Meaning: Integer arithmetic shift right.                                                |
' `-----------------------------------------------------------------------------------------------------'

Opc_ISHR        call    #PopRhsThenLhs
                sar     lhs,rhs
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             IUSHR                                                                               $37 |
' |             Meaning: Integer logical shift right.                                                   |
' `-----------------------------------------------------------------------------------------------------'

Opc_IUSHR       call    #PopRhsThenLhs
                shr     lhs,rhs
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             IAND                                                                                $38 |
' |             Meaning: Integer boolean AND.                                                           |
' `-----------------------------------------------------------------------------------------------------'

Opc_IAND        call    #PopRhsThenLhs
                and     lhs,rhs
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             IXOR                                                                                $39 |
' |             Meaning: Integer boolean XOR.                                                           |
' `-----------------------------------------------------------------------------------------------------'

Opc_IXOR        call    #PopRhsThenLhs
                xor     lhs,rhs
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             I2B                                                                                 $3A |
' |             Meaning: Integer to signed byte.                                                        |
' `-----------------------------------------------------------------------------------------------------'

Opc_I2B         call    #PopLhs
                shl     lhs,#1 WC, NR
        IF_C    or      lhs,#$80
                shl     lhs,#24
                sar     lhs,#24
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             GOTO                                                                                $3B |
' |             Meaning: Branch.                                                                        |
' |             Description: Execution proceeds at the signed 16-bit offset from the address            |
' |             of this instruction. The offset is constructed from byte1 and byte2.                    |
' `-----------------------------------------------------------------------------------------------------'

Opc_GOTO        jmp     #Branch

' .-----------------------------------------------------------------------------------------------------.
' |             INVOKEINTERFACE                                                                     $3C |
' |             Meaning: Invoke interface method.                                                       |
' |             Note                                                                                    |
' |             Unlike invokevirtual and invokespecial, the method block does not indicate the          |
' |             number of available arguments. This number is taken from the bytecode.                  |
' `-----------------------------------------------------------------------------------------------------'

Opc_INVOKEINTERFACE
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             ARRAYLENGTH                                                                         $3D |
' |             Meaning: Get length of array.                                                           |
' |             Description: The length of the array is determined and replaces aref on the             |
' |             top of the stack. The aref must be a reference to an array object.                      |
' |             Note                                                                                    |
' |             A NullPointerException is thrown if the aref is null.                                   |
' `-----------------------------------------------------------------------------------------------------'

Opc_ARRAYLENGTH
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             ATHROW                                                                              $3E |
' |             Meaning: Throw exception or error.                                                      |
' `-----------------------------------------------------------------------------------------------------'

Opc_ATHROW
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             CHECKCAST                                                                           $3F |
' |             Meaning: Make sure object is of given type.                                             |
' |             Description: Determines whether the referenced object can be cast to be a               |
' |             reference to an object of another class. A null object reference can be cast to         |
' |             any class. Otherwise, the referenced object must be an instance of the expected         |
' |             class or one of its superclasses. byte1 and byte2 are used to construct an              |
' |             index into the constant pool of the current class, which is presumed to be a            |
' |             class name that can be resolved to a class pointer.                                     |
' |             Note                                                                                    |
' |             A ClassCastException is thrown if the referenced object cannot be cast to the           |
' |             expected class.                                                                         |
' `-----------------------------------------------------------------------------------------------------'

Opc_CHECKCAST
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             INSTANCEOF                                                                          $40 |
' |             Meaning: Determine if an object is of given type and return result.                     |
' |             Description: Determines whether the referenced object can be cast to be a               |
' |             reference to an object of the expected class. This instruction will overwrite           |
' |             the object reference with 1 if the object is an instance of the expected class          |
' |             or one of its superclasses. Otherwise, the object reference is overwritten by 0.        |
' `-----------------------------------------------------------------------------------------------------'

Opc_INSTANCEOF
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             WIDE                                                                                $41 |
' `-----------------------------------------------------------------------------------------------------'

Opc_WIDE
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             IINC                                                                                $42 |
' |             Meaning: Increment local variable by constant.                                          |
' |             Description: Local variable at byte1 in the current Java frame must contain an          |
' |             integer. Its value is incremented by the value byte2, where byte2 is treated as         |
' |             a signed 8-bit quantity.                                                                |
' `-----------------------------------------------------------------------------------------------------'

Opc_IINC        rdbyte  opc,pc
                add     pc,#1
                shl     opc,#2
                add     opc,bp
                rdlong  lhs,opc

                rdbyte  rhs,pc
                add     pc,#1
                shl     rhs,#24
                sar     rhs,#24

                adds    lhs,rhs

                wrlong  lhs,opc

                jmp     #Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             GETFIELD                                                                            $43 |
' |             Meaning: Get 16bit field from object.                                                   |
' `-----------------------------------------------------------------------------------------------------'

Opc_GETFIELD
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             PUTFIELD                                                                            $44 |
' |             Meaning: Set 16bit field in object.                                                     |
' `-----------------------------------------------------------------------------------------------------'

Opc_PUTFIELD
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             NEW                                                                                 $45 |
' |             Meaning: Create new object.                                                             |
' |             Description: byte1 and byte2 are used to construct an index into the constant           |
' |             pool of the current class, which must be a class name that can be resolved to a         |
' |             class pointer. A new instance of that class is then created, and a reference to         |
' |             the object is pushed on the stack.                                                      |
' `-----------------------------------------------------------------------------------------------------'

Opc_NEW
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             NEWARRAY                                                                            $46 |
' |             Meaning: Allocate new array.                                                            |
' |             Description: A new array of a specific array type, capable of holding size              |
' |             elements, is allocated. The result is a reference to the new object. Allocation         |
' |             of an array large enough to contain size items of the specific array type is            |
' |             attempted and all elements of the array are initialized to 0.                           |
' |             size represents the number of elements in the new array and must be an                  |
' |             integer. The result is stored with an internal code that indicates the type of          |
' |             array to allocate. Possible values for the type of array are as follows:                |
' |             T_BOOLEAN(4), T_chAR(5), T_FLOAT(6), T_DOUBLE(7), T_BYTE(8), T_SHORT(9),                |
' |             T_INT(10), and T_LONG(11).                                                              |
' |             Note                                                                                    |
' |             A NegativeArraySizeException is thrown if size is less than 0. An                       |
' |             OutOfMemoryError is thrown if there is not enough memory to allocate the array.         |
' `-----------------------------------------------------------------------------------------------------'

Opc_NEWARRAY
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             ANEWARRAY                                                                           $47 |
' |             Meaning: Allocate new array of objects.                                                 |
' |             Description: A new array of the indicated class type and capable of holding             |
' |             size elements is allocated. The result is a reference to the new object.                |
' |             Allocation of an array large enough to contain size elements of the given class         |
' |             type is attempted and all elements of the array are initialized to null.                |
' |             size represents the number of elements in the new array and must be an                  |
' |             integer. byte1 and byte2 are used to construct an index into the constant pool          |
' |             of the current class. When the item at that index is resolved, the resulting            |
' |             entry must be a class.                                                                  |
' |             The anewarray instruction is used to create a single-dimension array.                   |
' |             Note                                                                                    |
' |             A NegativeArraySizeException is thrown if size is less than 0. An                       |
' |             OutOfMemoryError is thrown if there is not enough memory to allocate the array.         |
' `-----------------------------------------------------------------------------------------------------'

Opc_ANEWARRAY
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             IALOAD                                                                              $48 |
' |             Meaning: Load integer from array.                                                       |
' |             Description: The integer value at the array index is retrieved and pushed onto          |
' |             the top of the stack. The aref must be a reference to an array of integers;             |
' |             likewise, the index into the array must be an integer.                                  |
' |             Note                                                                                    |
' |             A NullPointerException is thrown if aref is null. An                                    |
' |             ArrayIndexOutOfBoundsException is thrown if the array index is not within the           |
' |             bounds of the array.                                                                    |
' `-----------------------------------------------------------------------------------------------------'

Opc_IALOAD
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             IASTORE                                                                             $49 |
' |             Meaning: Store into integer array.                                                      |
' |             Description: An integer value is popped off the stack and stored in the array           |
' |             at the index. The aref must be a reference to an array of integer values, and           |
' |             the index into the array must be an integer as well.                                    |
' `-----------------------------------------------------------------------------------------------------'

Opc_IASTORE
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             BALOAD                                                                              $4A |
' |             Meaning: Load signed byte from array.                                                   |
' |             Description: The signed byte value at the array index is retrieved, expanded            |
' |             to an integer, and pushed onto the top of the stack. The aref must be a                 |
' |             reference to an array of signed byte values, and the index into the array must          |
' |             be an integer.                                                                          |
' `-----------------------------------------------------------------------------------------------------'

Opc_BALOAD
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             BASTORE                                                                             $4B |
' |             Meaning: Store into signed byte array.                                                  |
' |             Description: An integer is popped off the stack, converted to a signed byte,            |
' |             and stored in the array at the index. The arrayref must be a reference to an            |
' |             array of signed bytes, and the index into the array must be an integer. If the          |
' |             integer value is too large to be a signed byte, it is truncated.                        |
' `-----------------------------------------------------------------------------------------------------'

Opc_BASTORE
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             LDC                                                                                 $4C |
' |             Meaning: Push item from constant pool.                                                  |
' |             Description: byte1 is used as an unsigned 8-bit index into the constant pool            |
' |             of the current class. After the item at that index is resolved, it is pushed            |
' |             onto the stack.                                                                         |
' |             Note                                                                                    |
' |             This bytecode instruction throws an OutOfMemoryError if a String object is              |
' |             being pushed and there is not enough memory to allocate space for it.                   |
' `-----------------------------------------------------------------------------------------------------'

Opc_LDC
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             LDC_W                                                                               $4D |
' |             Meaning: Push 2-byte item from constant pool.                                           |
' |             Description: byte1 and byte2 are used to construct an unsigned 16-bit index             |
' |             into the constant pool of the current class. After the item at that index is            |
' |             resolved, it is pushed onto the stack.                                                  |
' |             Note                                                                                    |
' |             This bytecode instruction throws an OutOfMemoryError if a String object is              |
' |             being pushed and there is not enough memory to allocate space for it.                   |
' `-----------------------------------------------------------------------------------------------------'

Opc_LDC_W
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             LDC2_W                                                                              $4E |
' |             Meaning: Push long or double from constant pool.                                        |
' |             Description: byte1 and byte2 are used to construct an unsigned 16-bit index             |
' |             into the constant pool of the current class. After the two-word constant at             |
' |             that index is resolved, it is pushed onto the stack.                                    |
' |             Note                                                                                    |
' |             This bytecode instruction throws an OutOfMemoryError if a String object is              |
' |             being pushed and there is not enough memory to allocate space for it.                   |
' `-----------------------------------------------------------------------------------------------------'

Opc_LDC2_W
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             IMUL                                                                                $4F |
' |             Meaning: Integer multiply.                                                              |
' `-----------------------------------------------------------------------------------------------------'

Opc_IMUL        call    #PopRhsThenLhs
                mov     acc,#0 
                          
                neg     lhs,lhs WC
                muxc    signBit,#1

                neg     rhs,rhs WC
        IF_C    xor     signBit,#1

Opc_IMUL_Loop   shr     rhs,#1 WC
     
        IF_C    add     acc,lhs

                shl     lhs,#1

                long    LMM | LMM_Tjnz | rhs << 9
                long    @Opc_IMUL_Loop

                mov     lhs,acc

                shr     signBit,#1 WC
        IF_C    neg     lhs,lhs

                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             IDIV                                                                                $50 |
' |             Meaning: Integer divide.                                                                |
' `-----------------------------------------------------------------------------------------------------'

Opc_IDIV
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             IREM                                                                                $51 |
' |             Meaning: Integer remainder.                                                             |
' `-----------------------------------------------------------------------------------------------------'

Opc_IREM
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             JSR                                                                                 $52 |
' |             Meaning: Jump subroutine.                                                               |
' |             Description: Execution proceeds at the signed 16-bit offset from the address            |
' |             of this instruction. The offset is constructed from byte1 and byte2. The                |
' |             address of the instruction immediately following the current instruction is             |
' |             pushed onto the stack to provide a return address from the subroutine.                  |
' `-----------------------------------------------------------------------------------------------------'

Opc_JSR         mov     lhs,pc
                call    #PushLhs
                jmp     #Branch

' .-----------------------------------------------------------------------------------------------------.
' |             RET                                                                                 $53 |
' |             Meaning: Return from subroutine.                                                        |
' |             Description: Local variable in the current Java frame must contain a return             |
' |             address. The contents of the local variable are written into the program                |
' |             counter.                                                                                |
' `-----------------------------------------------------------------------------------------------------'

Opc_RET         rdbyte  opc,pc
                add     pc,#1
                shl     opc,#2
                add     opc,bp
                rdlong  pc,opc
                jmp     #Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             TABLESWITCH                                                                         $54 |
' |             Meaning: Access jump table by index and jump.                                           |
' |             Description: Immediately after the tableswitch instruction, padding consisting          |
' |             of 0-3 zeros is inserted so that the next byte begins at an address that is a           |
' |             multiple of four. A series of offsets follows the padding. These offsets are            |
' |             signed 4-byte quantities and consist of a default offset, a low offset, and a           |
' |             high offset, followed by additional high - low + 1 offsets. These additional            |
' |             offsets are treated as a 0-based jump table.                                            |
' |             The index the offsets point to must be an integer. If the index is less than            |
' |             the low offset or greater than the high offset, the default offset is added to          |
' |             the address of the current instruction. Otherwise, the low offset is subtracted         |
' |             from the index, and the element at the position index - low offset in the jump          |
' |             table is extracted and added to the address of the current instruction.                 |
' `-----------------------------------------------------------------------------------------------------'

Opc_TABLESWITCH
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             LOOKUPSWITCH                                                                        $55 |
' |             Meaning: Access jump table by key match and jump.                                       |
' |             Description: Immediately after the lookupswitch instruction, padding                    |
' |             consisting of 0-3 zeros is                                                              |
' |             inserted so that the next byte begins at an address that is a multiple of               |
' |             four. A series of pairs of offsets follows the padding. These offsets are               |
' |             signed 4-byte quantities. The first item in the pair is the default offset, and         |
' |             the second item in the pair gives the number of pairs that follow. The                  |
' |             additional pairs consist of a match and an offset.                                      |
' |             The integer key on the stack is then compared against each of the matches. If           |
' |             the key is equal to one of the matches, the offset is added to the address of           |
' |             the current instruction. If the key does not match any of the matches, the              |
' |             default offset is added to the address of the current instruction.                      |
' `-----------------------------------------------------------------------------------------------------'

Opc_LOOKUPSWITCH
                jmp     #HandleWithSpin
                jmp     #PushLhsThenFetch

' .-----------------------------------------------------------------------------------------------------.
' |             End of LMM Code                                                                         |
' `-----------------------------------------------------------------------------------------------------'

LmmCodeEnd
            
